#include "sca.h"

#define MIN(a,b) ((a)<(b)?(a):(b))


#define INCMOD(a,m)   {(a)++;    if    ((a)>=(m)) {(a)=0;}}
#define ADDMOD(a,b,m) {(a)+=(b); while ((a)>=(m)) {(a)-=(m);}}

#define ComplementChar(a) ((a)^(0x55))


string reverseMotif(string& m) {
  string result = "";
  for(int i = m.size()-1 ; i >= 0 ; i--)
    result = result + m[i];
  return result;
}



sca::sca(string& motif, bool reversed){
  
  int nbshift = 0;
  _reversed = reversed;
  /* (1) Init seed object */
  _motif            = _reversed?reverseMotif(motif):motif;
  _motif_span       = 0;
  _motif_bitweight  = 0;
  
  // span of the seed is the size of the string
  _motif_span = _motif.size(); // FIXME -- this is not always true !! 
  
  /* (2) Fill _auto_shift table */
  for(int i=0 ; i< _motif_span + 3 ; i++){
    
    /* (2.1) increase if match/semimatch found */
    if (i < _motif_span ){
      switch(_motif[i]){
      case '#':
	nbshift+=2;
	_motif_bitweight+=2;
	break;
	
      case '@':
	nbshift+=1;
	_motif_bitweight+=1;
	break;
      }
    }
    
    /* (2.2) decrease if  match/semimatch left */
    if(i - 4 >= 0){
      switch(_motif[i-4]){
      case '#':
	nbshift-=2;
	break;
	
      case '@':
	nbshift-=1;
	break;
      }
    }
    _auto_shift[i] = nbshift; 
  }
  
  if (_motif_bitweight > (int)(8*sizeof(SEEDCODE))){
    cerr << "*ERROR : seed weight is to large to fit inside one word" << endl;
    exit(-1);
  }
  
      
   
  /* (3) Fill _auto_or table */

  // for all possible bytes (4 nucleotide in reverse order)
  for(int alpha = 0 ; alpha < 256 ; alpha++){ 
    
    // for each starting position on the seed
    for(int i=0;i<(int)_motif.length()+3;i++){
      
      int code = 0;

      // for each group of 4 nucleotides
      for(int j = 3; j >= 0; j--){
	// fill each box 
	if (i-j >= 0 && i-j < (int)_motif.length() )
	  switch(_motif[i-j]){
	  case '#':
	    code <<= 2;
	    code |=  ((_reversed?ComplementChar(alpha):alpha) >> (j<<1)) & 0x3;//...
	    break;
	  case '@':
	    code <<= 1;
	    code |=  ((_reversed?ComplementChar(alpha):alpha) >> (j<<1)) & 0x1;//...
	    break;
	  }//switch
      }//j			
      _auto_or[i][alpha] = code;
      //cout << "alpha : " << hex << alpha << ", i : " << i << " => " << code << endl;
      
    }//i		
  }//alpha	

  /* (4) .. well thats all for the moment .. */

}


int sca::init(seqdna& seqdna,int pos ){
  
  /* (1) init all vectors */
  _auto_nbboxes     = _motif_span + 3;
  _seqdnapos        =  0;
  _seqbytepos       = pos;
  _auto_fu          =  0;
  _seqdna           = seqdna;
  _auto_pos         = 4 - ((int)(_auto_nbboxes)&0x3);

#ifdef MASKED	
  _masked           = 0;
  _it_m             = _seqdna.maskpos.begin();
#endif  

  for(int i=0; i < _auto_nbboxes  ; i++)
    _auto_states[i] = 0;
  
  for(int j=1 ; j <= _auto_nbboxes >> 2 ; j++){
    int r = j << 2 ;
    for(int i=0 ; i < _auto_nbboxes && r > 0 ; i++){
      r--;
      _auto_states[i] <<=             _auto_shift[r];
      _auto_states[i]  |=  (SEEDCODE) _auto_or[r][seqdna.data[_seqbytepos]];
    }
    _seqbytepos++;
  }
  return 0;
}



SEEDCODE sca::next()
{ 
  SEEDCODE s = _auto_states[_auto_fu];

  /* (1) read last updated codes */  
 
  /* INCMOD(_auto_fu,_auto_nbboxes); */
  (++_auto_fu) %= _auto_nbboxes;
  
  /* (2) if the next element need update */
  if (_auto_fu == _auto_pos ){
    
    /*
    cout << "---------  " << dec << _auto_pos << endl;
    for(int i=0;i < _auto_nbboxes ; i++)
      cout << dec << i << ":" << hex << _auto_states[i]<< endl;
      cout << "---------  " << dec << _auto_pos << endl;
    */

    // no other elements
    int state_pos = _auto_pos;

    for(int i= _auto_nbboxes - 1 ; i >= 0 ; i--){
      /* (a) (clear new elements at the end of the table) */
      if (i < 4)
	_auto_states[state_pos] = 0;
      

      /* (b) (compute current element)  */
      _auto_states[state_pos] <<=  _auto_shift[i];
      _auto_states[state_pos] |=  (SEEDCODE) _auto_or[i][_seqdna.data[_seqbytepos]];
    


      /* INCMOD(state_pos,_auto_nbboxes); */
      (++state_pos) %= _auto_nbboxes;
      
      //cout << hex  << (int) _auto_shift[i] <<endl;
      //cout <<" _	auto_states[ i = " << dec << i << " ] = "<< hex << (int)_auto_states[i] << dec <<  endl;
    }
    _seqbytepos++;

    /* ADDMOD(_auto_pos,4,_auto_nbboxes); */
    _auto_pos+=4; _auto_pos %= _auto_nbboxes;
    
    
    /*
    cout << "=========  " <<  endl;
    for(int i=0;i < _auto_nbboxes ; i++)
      cout << dec << i << ":" << hex << _auto_states[i]<< endl;
    cout << "=========  " << endl;
    */
  }
  _seqdnapos++;

#ifdef MASKED	
  while (_it_m != _seqdna.maskpos.end() && (*(_it_m)) < _seqdnapos ) {
    (++_masked) %= 2;
    _it_m++;
  }
#endif	
  
  if (_reversed) {
    SEEDCODE rs = 0;
    for(int i=0 ; i<_motif_bitweight;i+=2) {
      SEEDCODE nucl = s & 0x3;
      s >>= 2;
      rs <<= 2;
      rs |= nucl;
    }
    //cerr << hex << s << ","<< rs << endl;
    return rs;
    
  } else {
    return s;
  }
}




std::ostream& operator<<(std::ostream& os, sca& sca){
  for(int alpha = 0; alpha < 256; alpha++){
    for(int i=0;i<12;i++){
      os << " _auto_or[ i = " << dec << i << " , alpha = " << hex << alpha <<"] = "<< hex << (int)(sca._auto_or[i][alpha]) <<  endl;
    }
  }
  return os;
}

  
